package perfSONAR::BWCTL;

use strict;
use warnings;

use IPC::Run qw( run timeout start finish pump);
use IO::Pty;

use perfSONAR qw( print_log);

#DEBUG
use Data::Dumper;
#DEBUG

#TODO Error handling
#composing of commandline and parsing of results of bwctl

sub commandline {
  my %parameters = @_;
  my @commandline;
  unless ($parameters{"src"} || $parameters{"dst"}) {
    return "ERROR", "Neither source nor destination ip address specified.";
  }
  if (!$parameters{"src"} || !$parameters{"dst"}){ #find out end point
    #my @result = `/sbin/ifconfig eth0`;
    #grep(/^\s+\w+\s+\w+\:(\d+\.\d+\.\d+\.\d+)/, @result);
    my $ip_result = `/sbin/ip addr show dev eth0 scope global primary`;
    $ip_result =~ /\n\s+inet\s+(\d+\.\d+\.\d+\.\d+)/;
    my $local_ip = $1;
    unless ($local_ip) {
      return "ERROR", "Cannot determine local ip address.";
    }
    $parameters{"src"} = $local_ip unless $parameters{"src"};
    $parameters{"dst"} = $local_ip unless $parameters{"dst"};
    if ($parameters{"src"} eq $parameters{"dst"}) {
      return "ERROR", "Measurement requested " .
        "using same local IP address as source and destination.";
    }
  }
  if ($parameters{"src"} eq $parameters{"dst"}) {
    return "ERROR", "Source ip address equal to destination ip address.";
  }

  #check parameters to be correct input:
  foreach my $param (keys %parameters){
    next if ($param eq "param_ns_prefix" ||
             $param eq "metaID" ||
             $param eq "subject_ns_prefix" ||
             $param eq "parameter_ID" ||
             $param eq "param_ns_uri" ||
             $param eq "subject_ns_uri");

    next if ($param eq "src" || $param eq "dst");
    if ($param eq "login" || $param eq "password"){
      unless ($parameters{$param} =~ /^\w+$/){ 
        return "ERROR", "Invalid login/password string specified.";
      }
      next;
    }
    if ($param eq "TOS"){
      unless ($parameters{$param} =~ /^\d+|0x\d+$/){ 
        return "ERROR", "Invalid TOS string specified.";
      }
      next;
    }
    if ($param eq "protocol"){
      unless ($parameters{$param} =~ /^udp$/i || $parameters{$param} =~ /^tcp$/i){
        return "ERROR", "Unknown protocol: $parameters{protocol}";
      }
      next;
    }
    unless ($parameters{$param} =~ /^\d+$/){
      return "ERROR", "Invalid value specified for $param.";
    }

  }    

  push @commandline , "-s" , $parameters{"src"}; 
  push @commandline , "AE", "AESKEY" if($parameters{"login"});
  push @commandline , $parameters{login} if($parameters{"login"});
  push @commandline , "-c", $parameters{dst};
  push @commandline , "AE", "AESKEY" if($parameters{"login"});
  push @commandline , $parameters{login} if($parameters{"login"});

  
  push @commandline , "-i", $parameters{interval} if($parameters{"interval"});
  push @commandline , "-t", $parameters{duration} if($parameters{"duration"});
  push @commandline , "-w", $parameters{windowSize} if($parameters{"windowSize"});

=cut obsolete
  if (! $parameters{"protocol"}) {
    # We assume default for undef, "" and 0
    # => Do nothing
  } elsif ($parameters{"protocol"} =~ /^udp$/i) {
    push @commandline , "-u";
  } elsif ($parameters{"protocol"} =~ /^tcp$/i) {
    # tcp is default
  } else {
    return "ERROR", "Unknown protocol: $parameters{protocol}";
  }
=cut

  push @commandline , "-u" if ($parameters{"protocol"} && $parameters{"protocol"} =~ /^udp$/i);
  push @commandline , "-l", $parameters{bufferSize} if($parameters{"bufferSize"});
  push @commandline , "-b", $parameters{bandwidth} if($parameters{"bandwidth"});
  push @commandline , "-S", $parameters{TOS} if($parameters{"TOS"});
  return @commandline;
}

sub do_bwctl {
  my $tool = shift;
  my $pass = shift;
  my @bwctl = @_;

  print_log ("info", "do_bwctl called with $tool\n");
  my ($in, $out, $err);
  unshift @bwctl, $tool;
  
  if(!$pass){
    my $command = join (" ", @bwctl);
    my $out=`$command 2>&1`;
    return $out;
  }
  my $h = start (\@bwctl, '<pty<', \$in, '>pty>', \$out, '2>', \$err);
  while (1){
    pump $h;
    if ($err =~ /passphrase/){
      $in = "$pass\n" ;
      $err = "";
    }
    elsif ($err ne "\n"){
      last;
    }
  }
    
  my $out_tmp = $out;
  my $err_tmp = $err;

  finish $h;
  if (!$out){
    return "$err_tmp" . "$err";
  }
  return "$out_tmp" . "$out";

}
sub parse_result {
  my $time = shift;
  my $result = shift;
  my @result = split(/\n/, $result);
  my @datalines;

  foreach my $resultline (@result){
    next unless  ($resultline =~
        /(\d+\.\d+\s*\-\s*\d+\.\d+)\s+sec\s+(\d+\.?\d*)\s+(\w+)\s+(\d+\.?\d*)\s+(\w+\/\w+)/);

    my %data_hash;
    $data_hash{"timeType"} = "unix";
    $data_hash{"timeValue"} = $time;
    $data_hash{"interval"} = $1;
    $data_hash{"numBytes"} = $2;
    $data_hash{"numBytesUnits"} = $3;
    $data_hash{"value"} = $4;
    $data_hash{"valueUnits"} = $5;
    push @datalines, \%data_hash;

  }
  if($#datalines < 0){
    #no data -> something wrong, write result as error description:
    $datalines[0]="BWCTL Error:";
    foreach my $resultline (@result){
      push @datalines, $resultline;
    }
  } 

  return @datalines;
} 


sub parse_header {
  my $result = shift;
  my @result = split(/\n/, $result);
  my %header;

  foreach my $resultline (@result){
    next if ($resultline =~
        /(\d+\.\d+\s*\-\s*\d+\.\d+)\s+sec\s+(\d+\.?\d*)\s+(\w+)\s+(\d+\.?\d*)\s+(\w+\/\w+)/);
    #print "header: $resultline\n";
    if ($resultline =~ /TCP/){
      $header{"proto"} = "TCP";
    }
    if ($resultline =~ /UDP/){
      $header{"proto"} = "UDP";
    }
    if ($resultline =~ /TCP window size:\s*(\d+)\s*(\w+)/){
      $header{"windowsize"} = $1; #$2 is always "Byte"
    }
    if ($resultline =~ /-t\s*(\d+)/){
      $header{"duration"} = $1;
    }
  }
  return %header;
}

1;
