package perfSONAR::Client::LS;

use strict;
use warnings;

#DEBUG
use Data::Dumper;
#/DEBUG


BEGIN {
  use vars qw($VERSION @ISA @EXPORT);
  use Exporter;
  @ISA = qw(Exporter);

  # set the version for version checking
  $VERSION     = 0.10;
  # if using RCS/CVS, this may be preferred
  #$VERSION = sprintf "%d.%03d", q$Revision: 1.1 $ =~ /(\d+)/g;
  @EXPORT      = qw(heartbeat deregister);          # Symbols to autoexport (:DEFAULT tag)
}

use NMWG::Message;
#use perfSONAR qw(print_log);


my $register_template = "$FindBin::RealBin/../etc/LS_register.xml";
my $deregister_template = "$FindBin::RealBin/../etc/LS_deregister.xml";
my $keepalive_template = "$FindBin::RealBin/../etc/LS_keepalive.xml";

my $psservice = "http://ggf.org/ns/nmwg/tools/org/perfsonar/service/1.0/";
my $nmwg = "http://ggf.org/ns/nmwg/base/2.0/";
my $nmtb = "http://ogf.org/schema/network/topology/base/20070828/";
my $nmtl3 = "http://ogf.org/schema/network/topology/l3/20070828/";
my $nmwgr = "http://ggf.org/ns/nmwg/result/2.0/";

our %services = ();
our (
  $hostname, $port, $organization, $contact, $keepalive, $log,
);

our @ls_url = ();
my %messages;

my $debug_write = 0;

# handle initial registration as "first" heartbeat.
# heartbeat is started as own process running in parallel to the server routine
sub heartbeat {

  #first registration:
  foreach my $service (keys %services){
    create_register_message($service);
    send_registration($messages{$service}{"register_msg"}, $service);

  }
  #send keepalive messages for each service
  while(1){
    sleep ($keepalive); #wait $keepalive seconds
    foreach my $service (keys %messages){ #get services from global hash
      my $message;
      if (exists $messages{$service}{"keepalive_msg"}){
        foreach my $url (@ls_url){
          $message = $messages{$service}{"keepalive_msg"}->clone();
          $log->log(
            level => "notice", service => $service,
            message => "Sending keepalive for $service to $url"
          );
          my $response = perfSONAR::sendReceive(
            message => $message,
            uri => $url,
          );
          if ($response){
            write_message($response, "keepalive-response");
            #TODO What to do if keepalive fails? At the moment: Nothing. 
            my $eventtype = ($response->{dom}->getElementsByTagNameNS("$nmwg", "eventType"))[0]->textContent;
            my $datumstring = ($response->{dom}->getElementsByTagNameNS("$nmwgr", "datum"))[0]->textContent;
            $log->log(
              level => "notice", service => $service,
              message => "Keepalive for $service returend $eventtype: $datumstring"
            );
          } else { #no response... 
            $log->log(
              level => "notice", service => $service,
              message => "error sending keepalive for $service: No response from Lookup Server!"
            );
          }
        }
      } else { #something went wrong with first registration, otherwise there would be a keepalive message!
               #try to send registration message again:
        send_registration($messages{$service}{"register_msg"}, $service);
      }
    }
  }
}


sub create_register_message{
  my $service = shift;

  # read in template for LS registration message and fill in module specific
  # values:
  my $message = NMWG::Message->new();
  $message->parse_xml_from_file($register_template);
  my $servicenode = ($message->{dom}->getElementsByTagNameNS("$psservice", "service"))[0];

  $message->add_element_NS($servicenode, "serviceName", "$services{$service}->{name}", $psservice);
  $message->add_element_NS($servicenode, "accessPoint", "http://$hostname:$port/services/$service", $psservice);
  $message->add_element_NS($servicenode, "serviceType", "$services{$service}->{servicetype}", $psservice);
  $message->add_element_NS($servicenode, "serviceDescription", "$services{$service}->{description}", $psservice);

  my $datanode = ($message->{dom}->getElementsByTagNameNS("$nmwg", "data"))[0];
  my $metanode = $message->add_attribute (namespace => $nmwg,
                                          parent => $datanode,
                                          nodename => "metadata",
                                          id => "topo-metadata",
                                         );
  my $subjectnode = $message->add_attribute (namespace => $nmwg,
                                            parent => $metanode,
                                            nodename => "subject",
                                            id => "topo-subject",
                                           );
  my @addresses = lookup_interfaces();

  my $nodenode = $message->add_attribute (namespace => $nmtb,
                                          prefix => "nmtb",
                                          parent => $subjectnode,
                                          nodename => "node",
                                         );
  foreach my $addr (@addresses) {
    my $name = lookup_hostname($addr);
    if ($name) {
      $message->add_attribute (namespace => $nmtb,
                               prefix => "nmtb",
                               parent => $nodenode,
                               nodename => "name",
                               type => "dns",
                               value => $name,
                              );
    }
  }
  foreach my $addr (@addresses) {
    my $addr_type;

    if ($addr =~ /:/) {
      $addr_type = "ipv6";
    } else {
      $addr_type = "ipv4";
    }

    my $portnode = $message->add_attribute (namespace => $nmtl3,
                                            prefix => "nmtl3",
                                            parent => $nodenode,
                                            nodename => "port",
                                           );
    $message->add_attribute (namespace => $nmtl3,
                             prefix => "nmtb",
                             parent => $portnode,
                             nodename => "address",
                             type => $addr_type,
                             value => $addr,
                            );
  }

  foreach my $et (@{$services{$service}{"eventtype"}}){
    $message->add_element_NS($metanode, "eventType", "$et", $nmwg);
  }
  my $paramnode = $message->add_attribute (namespace => $nmwg,
                                           parent => $metanode,
                                           nodename => "parameters",
                                           );
  if (exists $services{$service}->{"keyword"}){
    my $keywordnode = $message->add_attribute (namespace => $nmwg,
                                               parent => $paramnode,
                                               nodename =>"parameter",
                                               name => "keyword",
                                               value => $services{$service}->{keyword},
                                              );
  }


  if ($service =~/MA/){
    $services{$service}->{handler}->get_meta_info($message);
  }
  $messages{$service}{"register_msg"} = $message;
}

sub check_response {
  my $response = shift;
  my $service = shift;


  my ($errorstring, $metaid) = $response->parse_all;
  if($errorstring){
    $log->log(level => "notice", service => $service, message => $errorstring);
  }
  #parse content, write to log if successfull registration or error happened
  my $eventtype;
  my $key;
  foreach my $meta (keys %{$response->{"metadataIDs"}}){
    $eventtype = $response->{"metadataIDs"}{$meta}{"eventType"};
    if (!$eventtype =~ /success/){
      $log->log(
        level => "notice", service => $service,
        message => "LS returned $eventtype for service $service"
      );
      #return;
    } else {
      if ($eventtype =~ /register/){
        $key = $response->{"metadataIDs"}{$meta}{"key"}{"lsKey"};
        $log->log(
          level => "notice", service => $service,
          message => "successfully registered service $service with key $key"
        );
        $messages{$service}{"ls_key"} = $key;
      }
    }
  }
  return $key;
}

sub create_keepalive_message {
  my $key = shift;

  my $message = NMWG::Message->new();
  $message->parse_xml_from_file($keepalive_template);
  my $paramnode = ($message->{dom}->getElementsByTagNameNS("$nmwg", "parameters"))[0];
  $message->add_attribute(parent => $paramnode, nodename => "parameter", value => $key, name => "lsKey");
  write_message($message, "keepalive");
  return $message;
}  

sub send_registration {

  my $message = shift;
  my $service = shift;

  write_message($message, "register");
  foreach my $url (@ls_url){
    $log->log(
      level => "notice", service => $service,
      message => "registering service $service to $url"
    );
    my $response = perfSONAR::sendReceive(
          message => $message->clone,
          uri => $url,
    );
    if ($response){
      write_message($response, "register-response");
      my $key = check_response($response, $service);
      if ($key){
        $messages{$service}{"keepalive_msg"} = create_keepalive_message($key);
      } else {
        my $eventtype = ($response->{dom}->getElementsByTagNameNS("$nmwg", "eventType"))[0]->textContent;
        my $datumstring = ($response->{dom}->getElementsByTagNameNS("$nmwgr", "datum"))[0]->textContent;
        $log->log(
          level => "notice", service => $service,
          message => "error registering service $service with $eventtype: $datumstring"
        );
      }
    } else {
      $log->log(
        level => "notice", service => $service,
        message => "error registering service $service: No response from Lookup Server!"
      );
    }
  }
}



sub deregister{
  #read in deregistration template:
  my $message = NMWG::Message->new();
  $message->parse_xml_from_file($deregister_template);

  #send deregistration
  foreach my $service (keys %messages){
    my $key = $messages{$service}{"ls_key"};
    if (!$key){
      $log->log(
        level => "notice", service => $service,
        message => "no key for deregistering service $service found!"
      );
      next;
    }
    my $deregmsg = $message->clone;
    my $parametersnode = (
      $deregmsg->{dom}->getElementsByTagNameNS("$nmwg", "parameters")
    )[0];
    my $key_parameter = $deregmsg->add_attribute(
            parent => $parametersnode,
            nodename => "parameter",
            name => "lsKey"
    );
    $key_parameter->appendText($key);
    write_message($deregmsg, "deregister");
    foreach my $url (@ls_url){
      $log->log(
        level => "notice", service => $service,
        message => "deregistering service $service from $url"
      );
      my $response = perfSONAR::sendReceive(
            message => $deregmsg->clone,
            uri => $url,
      );
      if ($response){
        write_message($response, "dereg-response");
        $log->log(
          level => "notice", service => $service,
          message => "successfully deregistered service $service with key $key"
        );
        #TODO do something interesting with result
        #$response->parse_meta;
        #$response->parse_data;
      } else {
        $log->log(
          level => "notice", service => $service,
          message => "could not deregister $key: No response from Looukup Server!"
        );
      }
    }
  }
  #lay down to die
}

sub get_key{
#TODO not implemented yet
}

sub re_register{
#TODO not implemented yet
}

sub lookup_interfaces {
    my @ret_interfaces = ();

    open(IFCONFIG, "/sbin/ifconfig |");
    my $is_eth = 0;
    while(<IFCONFIG>) {
        if (/Link encap:([^ ]+)/) {
            if (lc($1) eq "ethernet") {
                $is_eth = 1;
            } else {
                $is_eth = 0;
            }
        }
        next if (not $is_eth);

        if (/inet \w+:(\d+\.\d+\.\d+\.\d+)/) {
            push @ret_interfaces, $1;
        } elsif (/inet6 \w+: (\d*:[^\/ ]*)(\/\d+)? .*:Global/) {
            push @ret_interfaces, $1;
        }
    }
    close (IFCONFIG);
    return @ret_interfaces;
}

sub lookup_hostname {
    my $ip = shift;
    my @h = gethostbyaddr(pack('C4',split('\.',$ip)),2);
    return $h[0];
}

sub write_message{
  my $message = shift;
  my $name = shift;

  if ($debug_write){
    my $timestamp = time;
    open (FH, ">", "$name-$timestamp.xml");
    print FH $message->as_string();
    close FH;
    sleep(1);
  }
}


1;

